# flake8: noqa: E402
from syftr.evaluation.evaluator_factory import CorrectnessEvaluatorFactory
from syftr.event_loop import fix_asyncio

fix_asyncio()

import asyncio
import logging
import math
import random
import typing as T
from collections import Counter, defaultdict
from datetime import datetime, timezone

import anthropic
import azure
import google
import llama_index.core.instrumentation as instrument
import numpy as np
import openai
from aiolimiter import AsyncLimiter
from llama_index.core.bridge.pydantic import Field
from llama_index.core.evaluation import BaseEvaluator, CorrectnessEvaluator
from llama_index.core.evaluation.base import EvaluationResult
from llama_index.core.llms import CompletionResponse
from llama_index.core.prompts.mixin import PromptDictType
from llama_index.core.schema import MetadataMode, NodeWithScore
from optuna import TrialPruned
from rapidfuzz.fuzz import partial_ratio
from scipy import stats
from sklearn import metrics
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception,
    retry_if_exception_cause_type,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential_jitter,
)

from syftr import core, custom_metrics
from syftr.configuration import EVAL__RAISE_ON_EXCEPTION
from syftr.flows import Flow, JudgeFlow, RetrieverFlow
from syftr.helpers import get_exception_report
from syftr.instrumentation.tokens import LLMCallData
from syftr.pruning import CostPruner, ParetoPruner, RuntimePruner
from syftr.studies import AgentStudyConfig, SearchSpace, StudyConfig

logging.basicConfig(level=logging.WARNING)
log = logging.getLogger("evaluation")
log.setLevel(logging.INFO)

dispatcher = instrument.get_dispatcher()
CorrectnessEvaluator.evaluate = dispatcher.span(CorrectnessEvaluator.evaluate)  # type: ignore
CorrectnessEvaluator.aevaluate = dispatcher.span(CorrectnessEvaluator.aevaluate)  # type: ignore

RETRY_ATTEMPTS = 3
RETRY_MIN_SLEEP = 2
RETRY_MAX_SLEEP = 60

RETRIABLE_EXCEPTIONS = (
    anthropic.RateLimitError,
    anthropic.APITimeoutError,
    anthropic.APIConnectionError,
    anthropic.InternalServerError,
    openai.APIConnectionError,
    openai.APITimeoutError,
    openai.RateLimitError,
    openai.InternalServerError,
    google.api_core.exceptions.ResourceExhausted,
    google.api_core.exceptions.ServiceUnavailable,
    azure.core.exceptions.HttpResponseError,
    azure.core.exceptions.ServiceRequestError,
    azure.core.exceptions.ServiceResponseError,
)


class ExactMatchEvaluator(BaseEvaluator):
    """
    Evaluator that calculates exact match by comparing reference contexts
    with retrieved contexts.
    """

    async def aevaluate(
        self,
        query: T.Optional[str] = None,
        response: T.Optional[str] = None,
        contexts: T.Optional[T.Sequence[str]] = None,
        **kwargs: T.Any,
    ) -> EvaluationResult:
        """
        Evaluate exact match by computing the proportion of reference contexts
        that are present in the retrieved contexts.
        """
        reference = kwargs.get("reference")

        if not reference:
            raise ValueError("Reference contexts are empty.")
        if not contexts:
            raise ValueError("Retrieved contexts are empty.")

        matched = sum(any(ref in context for context in contexts) for ref in reference)
        recall = matched / len(reference) if reference else 0.0
        return EvaluationResult(
            passing=recall > 0,
            score=recall,
        )

    def evaluate(
        self,
        query: T.Optional[str] = None,
        response: T.Optional[str] = None,
        contexts: T.Optional[T.Sequence[str]] = None,
        **kwargs: T.Any,
    ) -> EvaluationResult:
        """
        Synchronous version of the evaluation method for compatibility with base class.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self.aevaluate(query, response, contexts, **kwargs)
        )

    def _get_prompts(self) -> PromptDictType:
        """Get prompts."""
        return {}

    def _update_prompts(self, prompts_dict: PromptDictType) -> None:
        """Update prompts."""
        pass


class FuzzyRecallEvaluator(BaseEvaluator):
    """
    Evaluator that calculates fuzzy recall by comparing reference contexts
    with retrieved contexts using partial_ratio from rapidfuzz.
    """

    def __init__(self, threshold: float = 90.0):
        self.threshold = threshold

    async def fuzzy_match_async(self, ref: str, doc: str) -> bool:
        return await asyncio.to_thread(partial_ratio, ref, doc) >= self.threshold

    async def fuzzy_contains_async(self, ref: str, docs: T.Sequence[str]) -> bool:
        tasks = [self.fuzzy_match_async(ref, doc) for doc in docs]
        for coro in asyncio.as_completed(tasks):
            if await coro:
                return True
        return False

    async def aevaluate(
        self,
        query: T.Optional[str] = None,
        response: T.Optional[str] = None,
        contexts: T.Optional[T.Sequence[str]] = None,
        **kwargs: T.Any,
    ) -> EvaluationResult:
        """
        Evaluate fuzzy recall by computing the proportion of reference contexts
        that have a fuzzy match in the retrieved contexts.
        """
        reference = kwargs.get("reference")

        if not reference:
            raise ValueError("Reference contexts are empty.")
        if not contexts:
            raise ValueError("Retrieved contexts are empty.")

        tasks = [self.fuzzy_contains_async(ref, contexts) for ref in reference]
        results = await asyncio.gather(*tasks)
        matched = sum(results)
        recall = matched / len(reference) if reference else 0.0
        return EvaluationResult(
            passing=recall > 0,
            score=recall,
        )

    def evaluate(
        self,
        query: T.Optional[str] = None,
        response: T.Optional[str] = None,
        contexts: T.Optional[T.Sequence[str]] = None,
        **kwargs: T.Any,
    ) -> EvaluationResult:
        """
        Synchronous version of the evaluation method for compatibility with base class.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self.aevaluate(query, response, contexts, **kwargs)
        )

    def _get_prompts(self) -> PromptDictType:
        """Get prompts."""
        return {}

    def _update_prompts(self, prompts_dict: PromptDictType) -> None:
        """Update prompts."""
        pass


class JudgeEvaluator(BaseEvaluator):
    """
    Evaluator that checks if the evaluation result (passing or not) matches the labeled value
    """

    async def aevaluate(
        self,
        query: str = "",
        evaluation_result: T.Optional[EvaluationResult] = None,
        label: bool | None = None,
        **kwargs: T.Any,
    ) -> EvaluationResult:
        assert evaluation_result is not None, (
            "JudgeEvaluator needs an evaluation_result"
        )
        assert label is not None, "JudgeEvaluator needs a label"

        passing = evaluation_result.passing == label

        return EvaluationResult(
            query=query,
            passing=passing,
            response=str(evaluation_result),
        )

    def evaluate(
        self,
        query: str = "",
        evaluation_result: T.Optional[EvaluationResult] = None,
        label: str | int | None = None,
        **kwargs: T.Any,
    ) -> EvaluationResult:
        """
        Synchronous version of the evaluation method for compatibility with base class.
        """
        assert evaluation_result is not None, (
            "JudgeEvaluator needs an evaluation_result"
        )
        assert label is not None, "JudgeEvaluator needs a label"

        label = bool(int(label))
        passing = evaluation_result.passing == label

        return EvaluationResult(
            query=query,
            passing=passing,
            response=str(evaluation_result),
        )

    def _get_prompts(self) -> PromptDictType:
        """Get prompts."""
        return {}

    def _update_prompts(self, prompts_dict: PromptDictType) -> None:
        """Update prompts."""
        pass


class SyftrEvaluationResult(EvaluationResult):
    class Config:
        arbitrary_types_allowed = True

    qa_pair: T.Optional[core.QAPair] = Field(default=None, description="Q&A pair")
    run_time: T.Optional[float] = Field(
        default=np.nan, description="Flow completion time"
    )
    generation_exception: T.Optional[Exception] = Field(
        default=None, description="Exception during generation"
    )
    evaluation_exception: T.Optional[Exception] = Field(
        default=None, description="Exception during evaluation"
    )
    llm_call_data: T.List[LLMCallData] = Field(
        default_factory=list,
        description="Token counts and latencies for all LLM calls made during flow",
    )
    retriever_context_length: T.Optional[float] = Field(
        default=None, description="Total length of retrieved contexts in tokens"
    )
    retriever_recall: T.Optional[float] = Field(
        default=None, description="Retriever recall score in [0, 1]"
    )
    response_text: T.Optional[str] = Field(
        default=None, description="Text of generated response."
    )


async def exception_catcher(
    func: T.Callable,
    return_values_on_exception: T.Tuple[T.Any, ...] | None,
    raise_on_exception: bool | None = EVAL__RAISE_ON_EXCEPTION,
    **kwargs,
):
    try:
        results = await func(**kwargs)
        if isinstance(results, tuple):
            return *results, None
        return results, None
    except RETRIABLE_EXCEPTIONS:
        raise
    except Exception as ex:
        if raise_on_exception:
            log.exception(
                ">>>>>>>>>>>>>>>>> EXCEPTION <<<<<<<<<<<<<<<<<<<<\n %s for input: %s",
                ex,
                kwargs,
            )
            raise ex
        log.exception(">>> %s for input: %s", ex, kwargs)
        ret = return_values_on_exception or ()
        ret = (ret,) if not isinstance(ret, tuple) else ret
        return *ret, ex


def should_retry(exception):
    """Add any specific tests for exceptions which should not be retried here."""
    if isinstance(exception, Exception) and "draft tokens, too large for model" in str(
        exception
    ):
        msg = f"Trial pruned due to HttpResponseError 3051: {exception.message}"
        log.warning(msg)
        raise TrialPruned(msg)
    return True


@retry(
    stop=stop_after_attempt(RETRY_ATTEMPTS),
    wait=wait_exponential_jitter(
        initial=RETRY_MIN_SLEEP, max=RETRY_MAX_SLEEP, exp_base=2, jitter=RETRY_MAX_SLEEP
    ),
    reraise=True,
    retry=(
        retry_if_exception_type(RETRIABLE_EXCEPTIONS)
        | retry_if_exception_cause_type(RETRIABLE_EXCEPTIONS)
    )
    & retry_if_exception(should_retry),
    before_sleep=before_sleep_log(log, logging.WARNING),
)
async def agenerate_pair(
    qa_pair: core.QAPair,
    flow: Flow,
    rate_limiter: AsyncLimiter,
    raise_on_exception: bool | None = EVAL__RAISE_ON_EXCEPTION,
) -> T.Tuple[CompletionResponse | None, float, T.List[LLMCallData], Exception | None]:
    """Get flow's answer to a question from an Q&A pair asynchronously."""
    # random wait to avoid thundering herd
    await asyncio.sleep(random.uniform(0.1, 0.5))
    async with rate_limiter:
        return await exception_catcher(
            func=flow.agenerate,
            return_values_on_exception=(False, np.nan, []),
            raise_on_exception=raise_on_exception,
            query=qa_pair.question,
        )


@retry(
    stop=stop_after_attempt(RETRY_ATTEMPTS),
    wait=wait_exponential_jitter(
        initial=RETRY_MIN_SLEEP, max=RETRY_MAX_SLEEP, exp_base=2, jitter=RETRY_MAX_SLEEP
    ),
    reraise=True,
    retry=(
        retry_if_exception_type(RETRIABLE_EXCEPTIONS)
        | retry_if_exception_cause_type(RETRIABLE_EXCEPTIONS)
    )
    & retry_if_exception(should_retry),
    before_sleep=before_sleep_log(log, logging.WARNING),
)
async def aevaluate_pair(
    qa_pair: core.QAPair,
    response: CompletionResponse,
    evaluator: BaseEvaluator,
    rate_limiter: AsyncLimiter,
    raise_on_exception: bool | None = EVAL__RAISE_ON_EXCEPTION,
) -> T.Tuple[EvaluationResult | None, Exception | None]:
    """Evaluate a flow response asynchronously."""
    async with rate_limiter:
        return await exception_catcher(
            func=evaluator.aevaluate,
            return_values_on_exception=(None,),
            raise_on_exception=raise_on_exception,
            query=qa_pair.question,
            response=response.text,
            reference=qa_pair.answer,
        )


async def _aeval_pair(
    qa_pair: core.QAPair,
    flow: Flow,
    evaluators: T.Sequence[BaseEvaluator],
    rate_limiter: AsyncLimiter,
    raise_on_exception: bool | None = EVAL__RAISE_ON_EXCEPTION,
) -> SyftrEvaluationResult:
    """Evaluate single Q&A item asynchronously."""
    response, run_time, call_data, generation_exception = await agenerate_pair(
        qa_pair, flow, rate_limiter
    )
    eval_result, evaluation_exception = None, None
    if response:
        evaluator = evaluators[0]
        eval_result, evaluation_exception = await aevaluate_pair(
            qa_pair, response, evaluator, rate_limiter, raise_on_exception
        )
    return SyftrEvaluationResult(
        qa_pair=qa_pair,
        run_time=run_time,
        generation_exception=generation_exception,
        evaluation_exception=evaluation_exception,
        llm_call_data=call_data,
        response_text=response.text,
        **(eval_result.model_dump() if eval_result else {}),
    )


@retry(
    stop=stop_after_attempt(RETRY_ATTEMPTS),
    wait=wait_exponential_jitter(
        initial=RETRY_MIN_SLEEP, max=RETRY_MAX_SLEEP, exp_base=2, jitter=RETRY_MAX_SLEEP
    ),
    reraise=True,
    retry=(
        retry_if_exception_type(RETRIABLE_EXCEPTIONS)
        | retry_if_exception_cause_type(RETRIABLE_EXCEPTIONS)
    )
    & retry_if_exception(should_retry),
    before_sleep=before_sleep_log(log, logging.WARNING),
)
async def aretrieve_pair(
    qa_pair: core.QAPair,
    flow: RetrieverFlow,
    rate_limiter: AsyncLimiter,
    raise_on_exception: bool | None = EVAL__RAISE_ON_EXCEPTION,
) -> T.Tuple[T.List[NodeWithScore] | None, float, Exception | None]:
    """Get flow's retrieved documents from an Q&A pair asynchronously."""
    result, run_time, exception = await exception_catcher(
        func=flow.aretrieve,
        return_values_on_exception=(None, np.nan),
        raise_on_exception=raise_on_exception,
        query=qa_pair.question,
    )
    return result, run_time, exception


async def _aeval_retriever_pair(
    qa_pair: core.QAPair,
    flow: RetrieverFlow,
    evaluators: T.Sequence[BaseEvaluator],
    rate_limiter: AsyncLimiter,
    raise_on_exception: bool | None = EVAL__RAISE_ON_EXCEPTION,
) -> SyftrEvaluationResult:
    """Evaluate retrieval performance on a single Q&A item."""
    if not qa_pair.gold_evidence:
        raise ValueError("QAPair gold_evidence is empty: %s", qa_pair)
    retrieval_results, run_time, retrieval_exception = await aretrieve_pair(
        qa_pair, flow, rate_limiter
    )
    if retrieval_results:
        retrieved_contexts: T.List[str] = [
            n.node.get_content(metadata_mode=MetadataMode.NONE)
            for n in retrieval_results or []
        ]
        # Approximate token length used here for speed of implementation.
        retrieved_contexts_length = sum(len(s) // 4 for s in retrieved_contexts)
        evaluator = evaluators[0]
        result = await evaluator.aevaluate(
            contexts=retrieved_contexts,
            reference=qa_pair.gold_evidence,
        )
    return SyftrEvaluationResult(
        qa_pair=qa_pair,
        run_time=run_time,
        generation_exception=retrieval_exception,
        evaluation_exception=None,
        llm_call_data=[],
        retriever_recall=result.score,
        retriever_context_length=retrieved_contexts_length,
        passing=result.passing,
    )


@retry(
    stop=stop_after_attempt(RETRY_ATTEMPTS),
    wait=wait_exponential_jitter(
        initial=RETRY_MIN_SLEEP, max=RETRY_MAX_SLEEP, exp_base=2, jitter=RETRY_MAX_SLEEP
    ),
    reraise=True,
    retry=(
        retry_if_exception_type(RETRIABLE_EXCEPTIONS)
        | retry_if_exception_cause_type(RETRIABLE_EXCEPTIONS)
    )
    & retry_if_exception(should_retry),
    before_sleep=before_sleep_log(log, logging.WARNING),
)
async def ajudge_pair(
    qa_pair: core.QAPair,
    flow: JudgeFlow,
    rate_limiter: AsyncLimiter,
    raise_on_exception: bool | None = EVAL__RAISE_ON_EXCEPTION,
) -> T.Tuple[EvaluationResult | None, float, T.List[LLMCallData], Exception | None]:
    """Get flow's retrieved documents from an Q&A pair asynchronously."""
    result, run_time, call_data, exception = await exception_catcher(
        func=flow.ajudge,
        return_values_on_exception=(None, np.nan, []),
        raise_on_exception=raise_on_exception,
        question=qa_pair.question,
        answer=qa_pair.answer,
        response=qa_pair.response,
    )
    return result, run_time, call_data, exception


async def _aeval_judge_pair(
    qa_pair: core.QAPair,
    flow: JudgeFlow,
    evaluators: T.Sequence[BaseEvaluator],
    rate_limiter: AsyncLimiter,
    raise_on_exception: bool | None = EVAL__RAISE_ON_EXCEPTION,
) -> SyftrEvaluationResult:
    """Evaluate retrieval performance on a single Q&A item."""
    judge_result, run_time, call_data, exception = await ajudge_pair(
        qa_pair, flow, rate_limiter
    )
    if judge_result:
        evaluator = evaluators[0]
        result = await evaluator.aevaluate(
            query=qa_pair.question,
            evaluation_result=judge_result,
            label=qa_pair.label,
        )
        return SyftrEvaluationResult(
            qa_pair=qa_pair,
            run_time=run_time,
            generation_exception=exception,
            evaluation_exception=None,
            llm_call_data=call_data,
            passing=result.passing,
        )
    else:
        return SyftrEvaluationResult(
            qa_pair=qa_pair,
            run_time=run_time,
            generation_exception=exception,
            evaluation_exception=None,
            llm_call_data=call_data,
        )


async def _aeval_pair_debias(
    qa_pair: core.QAPair,
    flow: Flow,
    evaluators: T.Sequence[BaseEvaluator],
    rate_limiter: AsyncLimiter,
    raise_on_exception: bool | None = EVAL__RAISE_ON_EXCEPTION,
) -> SyftrEvaluationResult:
    """Evaluate single Q&A item asynchronously with an evaluator chosen at random."""
    response, run_time, call_data, generation_exception = await agenerate_pair(
        qa_pair, flow, rate_limiter
    )
    eval_result, evaluation_exception = None, None
    if response:
        evaluator = random.choice(evaluators)
        eval_result, evaluation_exception = await aevaluate_pair(
            qa_pair, response, evaluator, rate_limiter, raise_on_exception
        )

    return SyftrEvaluationResult(
        qa_pair=qa_pair,
        run_time=run_time,
        generation_exception=generation_exception,
        evaluation_exception=evaluation_exception,
        llm_call_data=call_data,
        response_text=response.text,
        **(eval_result.model_dump() if eval_result else {}),
    )


async def _aeval_pair_consensus(
    qa_pair: core.QAPair,
    flow: Flow,
    evaluators: T.Sequence[BaseEvaluator],
    rate_limiter: AsyncLimiter,
    raise_on_exception: bool | None = EVAL__RAISE_ON_EXCEPTION,
) -> SyftrEvaluationResult:
    """Evaluate single Q&A item asynchronously using provided evaluators and average the results."""
    response, run_time, call_data, generation_exception = await agenerate_pair(
        qa_pair, flow, rate_limiter
    )
    eval_result, evaluation_exception = None, None
    if response is not None:
        eval_results: T.List[EvaluationResult] = []
        for evaluator in evaluators:
            eval_result, exception = await aevaluate_pair(
                qa_pair,
                response,
                evaluator,
                rate_limiter,
                raise_on_exception,
            )
            if not eval_result:
                evaluation_exception = exception
                continue
            eval_results.append(eval_result)
        eval_result = eval_results[0]
        eval_result.passing = Counter([r.passing for r in eval_results]).most_common(1)[
            0
        ][0]
    return SyftrEvaluationResult(
        qa_pair=qa_pair,
        run_time=run_time,
        generation_exception=generation_exception,
        evaluation_exception=evaluation_exception,
        llm_call_data=call_data,
        response_text=response.text,
        **(eval_result.model_dump() if eval_result else {}),
    )


async def _aeval_all_pair_runner(
    pair_eval_runner: T.Callable,
    dataset: T.List[core.QAPair],
    flow: Flow,
    evaluators: T.Sequence[BaseEvaluator],
    eval_timeout: int,
    rate_limiter: AsyncLimiter,
    raise_on_exception: bool | None = EVAL__RAISE_ON_EXCEPTION,
) -> T.List[SyftrEvaluationResult]:
    """Helper function to run multiple pair_eval_runners in parallel."""
    tasks = []
    for pair in dataset:
        tasks.append(
            asyncio.create_task(
                pair_eval_runner(
                    pair, flow, evaluators, rate_limiter, raise_on_exception
                )
            )
        )

    await asyncio.wait(tasks, timeout=eval_timeout)

    all_results = []
    for t in tasks:
        try:
            r = t.result()
        except asyncio.exceptions.InvalidStateError as exc:
            # Providing empty result for proper reporting.
            exc.add_note(
                f"Eval of task {t} terminated due to timeout of {eval_timeout} seconds."
            )
            r = SyftrEvaluationResult(
                qa_pair=None,
                run_time=np.nan,
                generation_exception=exc,
                llm_call_data=[],
            )
        all_results.append(r)
    return all_results


def _async_eval_runner(
    pair_eval_runner: T.Callable,
    items: T.List[core.QAPair],
    flow: Flow,
    evaluators: T.Sequence[BaseEvaluator],
    study_config: T.Union[StudyConfig, AgentStudyConfig],
    rate_limiter: AsyncLimiter,
    raise_on_exception: bool | None = EVAL__RAISE_ON_EXCEPTION,
    pruner: ParetoPruner | None = None,
    timeout_pruner: RuntimePruner | None = None,
    cost_pruner: CostPruner | None = None,
) -> T.Tuple[T.List[SyftrEvaluationResult], T.Optional[str]]:
    """Evaluate Q&A items asynchronously using provided pair_eval_runner."""
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    prune_reason = None
    results = []
    batch_size = study_config.optimization.num_eval_batch
    num_batches = math.ceil(len(items) / batch_size)
    batches = [items[i : i + batch_size] for i in range(0, len(items), batch_size)]
    for i, batch in enumerate(batches):
        batch_result = loop.run_until_complete(
            _aeval_all_pair_runner(
                pair_eval_runner,
                list(batch),
                flow,
                evaluators,
                study_config.timeouts.single_eval_timeout,
                rate_limiter,
                raise_on_exception,
            )
        )
        results.extend(batch_result)
        run_times = [
            r.run_time for r in results if r and r.run_time and not np.isnan(r.run_time)
        ]
        # Compute stats and pruners if we have successful evals
        # or, if we are over the max fail rate, proceed to calculate_metrics, which will
        # error out if we don't have any successful trials.
        # max_eval_failure_rate only applies if there are zero successful evals so far
        if (
            run_times
            or i / num_batches > study_config.optimization.max_eval_failure_rate
        ):
            current_metrics = calculate_metrics(results, study_config)

            log.info(
                "Finished evaluation batch %s/%s with %s QA pairs. Metrics: %s, Flow: %s",
                i + 1,
                num_batches,
                len(batch),
                {
                    current_metrics["objective_1_name"]: current_metrics["obj1_value"],
                    current_metrics["objective_2_name"]: current_metrics["obj2_value"],
                    "num_errors": current_metrics["num_errors"],
                },
                flow,
            )

            if timeout_pruner:
                timeout_pruner.report_and_raise_on_prune(
                    step=(i + 1) * study_config.optimization.num_eval_batch,
                    p80_time=current_metrics["p80_time"],
                )

            if cost_pruner:
                cost_pruner.report_and_raise_on_prune(
                    step=(i + 1) * study_config.optimization.num_eval_batch,
                    total_cost=current_metrics["llm_cost_total"],
                    llm_cost_mean=current_metrics["llm_cost_mean"],
                )
            if pruner:
                if pruner.report_and_prune(
                    step=(i + 1) * study_config.optimization.num_eval_batch,
                    obj1=current_metrics["obj1_value"],
                    obj2=current_metrics["obj2_value"],
                    obj1_confidence=current_metrics["obj1_confidence"],
                    obj2_confidence=current_metrics["obj2_confidence"],
                ):
                    prune_reason = "pareto"
                    break
    return results, prune_reason


def async_eval(
    items: T.List[core.QAPair],
    flow: Flow,
    study_config: T.Union[StudyConfig, AgentStudyConfig],
    evaluators: T.Sequence[BaseEvaluator],
    rate_limiter: AsyncLimiter,
    pruner: ParetoPruner | None = None,
    cost_pruner: CostPruner | None = None,
    timeout_pruner: RuntimePruner | None = None,
    raise_on_exception: bool | None = EVAL__RAISE_ON_EXCEPTION,
) -> T.Tuple[T.List[SyftrEvaluationResult], T.Optional[str]]:
    """Evaluate Q&A items asynchronously with an evaluator chosen at random for each pair."""
    return _async_eval_runner(
        _aeval_pair,
        items,
        flow,
        evaluators,
        study_config,
        rate_limiter,
        pruner=pruner,
        timeout_pruner=timeout_pruner,
        raise_on_exception=raise_on_exception,
        cost_pruner=cost_pruner,
    )


def async_eval_debias(
    items: T.List[core.QAPair],
    flow: Flow,
    study_config: T.Union[StudyConfig, AgentStudyConfig],
    evaluators: T.Sequence[BaseEvaluator],
    rate_limiter: AsyncLimiter,
    pruner: ParetoPruner | None = None,
    cost_pruner: CostPruner | None = None,
    timeout_pruner: RuntimePruner | None = None,
    raise_on_exception: bool | None = EVAL__RAISE_ON_EXCEPTION,
) -> T.Tuple[T.List[SyftrEvaluationResult], T.Optional[str]]:
    """Evaluate Q&A items asynchronously with an evaluator chosen at random for each pair."""
    return _async_eval_runner(
        _aeval_pair_debias,
        items,
        flow,
        evaluators,
        study_config,
        rate_limiter,
        raise_on_exception=raise_on_exception,
        pruner=pruner,
        timeout_pruner=timeout_pruner,
        cost_pruner=cost_pruner,
    )


def async_eval_consensus(
    items: T.List[core.QAPair],
    flow: Flow,
    study_config: T.Union[StudyConfig, AgentStudyConfig],
    evaluators: T.Sequence[BaseEvaluator],
    rate_limiter: AsyncLimiter,
    pruner: ParetoPruner | None = None,
    cost_pruner: CostPruner | None = None,
    timeout_pruner: RuntimePruner | None = None,
    raise_on_exception: bool | None = EVAL__RAISE_ON_EXCEPTION,
) -> T.Tuple[T.List[SyftrEvaluationResult], T.Optional[str]]:
    """Evaluate Q&A items asynchronously with multiple evaluators and result averaging."""
    return _async_eval_runner(
        _aeval_pair_consensus,
        items,
        flow,
        evaluators,
        study_config,
        rate_limiter,
        raise_on_exception=raise_on_exception,
        pruner=pruner,
        timeout_pruner=timeout_pruner,
        cost_pruner=cost_pruner,
    )


def validate_evaluation_data(results: T.List[SyftrEvaluationResult]):
    generation_exceptions = []
    evaluation_exceptions = []
    run_times = [
        r.run_time for r in results if r and r.run_time and not np.isnan(r.run_time)
    ]
    call_data = [result.llm_call_data for result in results]
    costs = [sum(call.cost for call in calls) for calls in call_data]

    for ex_field in "generation_exception", "evaluation_exception":
        for result in results:
            ex: Exception = getattr(result, ex_field)
            if ex:
                if ex_field == "generation_exception":
                    generation_exceptions.append(ex)
                else:
                    evaluation_exceptions.append(ex)

    if generation_exceptions or evaluation_exceptions:
        exceptions = []
        if generation_exceptions:
            exceptions.append(
                ExceptionGroup(
                    "Exceptions during generation",
                    generation_exceptions,
                )
            )
        if evaluation_exceptions:
            exceptions.append(
                ExceptionGroup(
                    "Exceptions during evaluation",
                    evaluation_exceptions,
                )
            )
        exception_group = ExceptionGroup("Trial has failed evals", exceptions)
        exception_message = get_exception_report(exception_group)
        log.warning(exception_message)
        if not run_times or not costs:
            raise Exception(exception_message)
    else:
        logging.info("The evaluation finished without exceptions")


def eval_dataset(
    study_config: T.Union[StudyConfig, AgentStudyConfig],
    dataset_iter,
    flow: Flow,
    evaluation_mode: T.Literal[
        "single", "random", "consensus", "retriever", "judge"
    ] = "single",
) -> T.Dict[str, T.Any]:
    eval_start = datetime.now(timezone.utc).timestamp()
    dataset = list(dataset_iter.iter_examples())
    assert len(dataset) > 2, dataset
    dataset = dataset[: study_config.optimization.num_eval_samples]
    assert evaluation_mode in {"single", "random", "consensus", "retriever", "judge"}, (
        "Evaluation mode should be 'single', 'random', 'consensus', 'retriever' or 'judge'."
    )

    evaluators = CorrectnessEvaluatorFactory(study_config).get_evaluators()
    rate_limiter = AsyncLimiter(
        study_config.optimization.rate_limiter_max_coros,
        study_config.optimization.rate_limiter_period,
    )
    timeout, pruner, costout = None, None, None
    if not getattr(flow, "enforce_full_evaluation", False):
        if study_config.optimization.use_pareto_pruner:
            pruner = ParetoPruner(
                study_name=study_config.name,
                num_warmup_steps=study_config.optimization.num_warmup_steps_pareto,
                success_rate=study_config.optimization.pareto_pruner_success_rate,
            )
        if study_config.optimization.use_cost_pruner:
            costout = CostPruner(
                num_warmup_steps=study_config.optimization.num_warmup_steps_costout,
                num_total_steps=len(dataset),
                max_cost=study_config.optimization.max_trial_cost,
            )
        if study_config.optimization.use_runtime_pruner:
            timeout = RuntimePruner(
                num_warmup_steps=study_config.optimization.num_warmup_steps_timeout,
                num_total_steps=len(dataset),
                eval_timeout=study_config.timeouts.eval_timeout,
            )

    results: T.List[SyftrEvaluationResult] = []
    match evaluation_mode:
        case "single":
            results, prune_reason = async_eval(
                dataset,
                flow,
                study_config,
                evaluators=evaluators,
                raise_on_exception=study_config.evaluation.raise_on_exception,
                pruner=pruner,
                cost_pruner=costout,
                timeout_pruner=timeout,
                rate_limiter=rate_limiter,
            )
        case "random":
            results, prune_reason = async_eval_debias(
                dataset,
                flow,
                study_config,
                evaluators=evaluators,
                raise_on_exception=study_config.evaluation.raise_on_exception,
                pruner=pruner,
                cost_pruner=costout,
                timeout_pruner=timeout,
                rate_limiter=rate_limiter,
            )
        case "consensus":
            results, prune_reason = async_eval_consensus(
                dataset,
                flow,
                study_config,
                evaluators=evaluators,
                raise_on_exception=study_config.evaluation.raise_on_exception,
                pruner=pruner,
                cost_pruner=costout,
                timeout_pruner=timeout,
                rate_limiter=rate_limiter,
            )
        case "retriever":
            # Filter the dataset to remove any qa_pairs without gold_evidence
            filtered_dataset = [pair for pair in dataset if pair.gold_evidence]
            if not filtered_dataset:
                raise ValueError("No QAPairs with gold_evidence found in the dataset.")
            results, prune_reason = _async_eval_runner(
                pair_eval_runner=_aeval_retriever_pair,
                items=filtered_dataset,
                flow=flow,
                study_config=study_config,
                evaluators=[FuzzyRecallEvaluator()],
                raise_on_exception=study_config.evaluation.raise_on_exception,
                pruner=pruner,
                timeout_pruner=timeout,
                cost_pruner=costout,
                rate_limiter=rate_limiter,
            )
        case "judge":
            results, prune_reason = _async_eval_runner(
                pair_eval_runner=_aeval_judge_pair,
                items=dataset,
                flow=flow,
                study_config=study_config,
                evaluators=[JudgeEvaluator()],
                raise_on_exception=study_config.evaluation.raise_on_exception,
                pruner=pruner,
                timeout_pruner=timeout,
                cost_pruner=costout,
                rate_limiter=rate_limiter,
            )

    metrics = calculate_metrics(results, study_config) if results else {}
    log.info("Number of evaluations: %d", metrics.get("num_total", 0))
    log.info("Number of successful evaluations: %d", metrics.get("num_success", 0))
    log.info("Number of errored evaluations: %d", metrics.get("num_errors", 0))
    eval_end = datetime.now(timezone.utc).timestamp()
    eval_duration = eval_end - eval_start
    metrics["is_pruned"] = prune_reason is not None
    metrics["eval_start"] = eval_start
    metrics["eval_end"] = eval_end
    metrics["eval_duration"] = eval_duration
    metrics["total_qa_pairs"] = len(dataset)
    metrics["prune_reason"] = prune_reason
    metrics["cardinality"] = (
        study_config.search_space.get_cardinality()
        if isinstance(study_config.search_space, SearchSpace)
        else None
    )
    return metrics


def calculate_metrics(
    results: T.List[SyftrEvaluationResult],
    study_config: T.Union[StudyConfig, AgentStudyConfig],
) -> T.Dict[str, T.Any]:
    validate_evaluation_data(results)
    objective_1: str = study_config.optimization.objective_1_name
    objective_2: str = study_config.optimization.objective_2_name
    if study_config.is_retriever_study:
        assert objective_1 == "retriever_recall"
        assert objective_2 == "retriever_context_length"

    passing = [r.passing for r in results if r.passing in [True, False]]

    if len(passing) / len(results) < study_config.evaluation.min_reporting_success_rate:
        raise TrialPruned(
            f"Too few successful evaluations: {len(passing)} out of {len(results)}"
        )

    num_total = len(passing)
    num_errors = sum(
        int(res.passing is None)
        or int(res.generation_exception is not None)
        or int(res.evaluation_exception is not None)
        for res in results
    )
    num_generation_errors = sum(
        int(res.generation_exception is not None) for res in results
    )
    num_evaluation_errors = sum(
        int(res.evaluation_exception is not None) for res in results
    )
    acc = sum(passing) / num_total
    passing_std = np.std(passing)

    f1_scores = [
        core.f1_score(result.qa_pair.answer, result.response)
        for result in results
        if result and result.qa_pair and result.response
    ]
    f1_score = np.mean(f1_scores)
    retriever_recalls = [
        r.retriever_recall for r in results if r.retriever_recall is not None
    ]
    mean_retriever_recall = (
        float(np.mean(retriever_recalls)) if retriever_recalls else 0.0
    )
    retriever_context_lengths = [
        r.retriever_context_length for r in results if r.retriever_context_length
    ]
    mean_retriever_context_length = (
        float(np.mean(retriever_context_lengths)) if retriever_context_lengths else 0.0
    )
    run_times = [
        r.run_time for r in results if r and r.run_time and not np.isnan(r.run_time)
    ]
    min_time = float(np.min(run_times))
    max_time = float(np.max(run_times))
    mean_time = float(np.mean(run_times))
    median_time = float(np.median(run_times))
    p80_time = float(np.percentile(run_times, 80))
    run_times_std = float(np.std(run_times))
    run_times_p80 = [r for r in run_times if r <= p80_time]
    run_times_p80_std = float(np.std(run_times_p80))

    latency_data = extract_llm_latency_data(results)
    cost_data = extract_cost_data(results)
    token_data = extract_token_data(results)
    eval_results = extract_eval_results(
        results, include_responses=study_config.optimization.include_responses
    )
    judge_data = extract_judge_results(results) if study_config.is_judge_study else {}

    if objective_1 == "accuracy":
        obj1_value = acc
    elif objective_1 == "retriever_recall":
        obj1_value = mean_retriever_recall
    else:
        raise ValueError(f"Unknown objective_1: {objective_1}.")
    obj1_confidence = custom_metrics.acc_confidence(
        accuracy=obj1_value,
        n_samples=len(passing),
        zscore=study_config.optimization.obj1_zscore,
    )
    if objective_2 == "p80_time":
        obj2_values = run_times
        obj2_value = p80_time
    elif objective_2 == "llm_cost_mean":
        call_data = [result.llm_call_data for result in results]
        costs = [sum(call.cost for call in calls) for calls in call_data]
        obj2_values = costs
        obj2_value = float(np.mean(costs))
    elif objective_2 == "retriever_context_length":
        obj2_values = retriever_context_lengths
        obj2_value = mean_retriever_context_length
    else:
        raise ValueError(
            f"Unknown objective_2: {objective_2}. Valid options are: p80_time, llm_cost_mean, retriever_context_length."
        )
    obj2_confidence = custom_metrics.lognormal_confidence(
        values=obj2_values,
        zscore=study_config.optimization.obj2_zscore,
    )
    return {
        objective_1: obj1_value,
        objective_2: obj2_value,
        "min_time": min_time,
        "max_time": max_time,
        "mean_time": mean_time,
        "median_time": median_time,
        "passing_std": passing_std,
        "f1_score": f1_score,
        "num_total": num_total,
        "num_errors": num_errors,
        "num_generation_errors": num_generation_errors,
        "num_evaluation_errors": num_evaluation_errors,
        "num_success": num_total - num_errors,
        "p80_time": p80_time,
        "run_times_std": run_times_std,
        "run_times_p80_std": run_times_p80_std,
        "obj1_value": obj1_value,
        "obj1_confidence": obj1_confidence,
        "obj2_value": obj2_value,
        "obj2_confidence": obj2_confidence,
        "acc_confidence": obj1_confidence,
        "objective_1_name": objective_1,
        "objective_2_name": objective_2,
        "eval_results": eval_results,
        **cost_data,
        **token_data,
        **latency_data,
        **judge_data,
    }


def extract_eval_results(
    all_results: T.List[SyftrEvaluationResult], include_responses: bool = False
) -> T.Dict:
    if include_responses:
        return {
            res.qa_pair.id: {
                "passing": 1 if res.passing else 0,
                "raw_score": res.score,
                "question": res.qa_pair.question,
                "answer": res.qa_pair.answer,
                "response": res.response_text,
            }
            for res in all_results
            if res.qa_pair is not None
        }
    else:
        return {
            res.qa_pair.id: {"passing": 1 if res.passing else 0, "raw_score": res.score}
            for res in all_results
            if res.qa_pair is not None
        }


def extract_llm_latency_data(
    all_results: T.List[SyftrEvaluationResult],
) -> T.Dict[str, float]:
    call_data = [result.llm_call_data for result in all_results]
    per_model_latency = defaultdict(list)
    for calls in call_data:
        for call in calls:
            per_model_latency[call.llm_name].append(call.llm_call_latency)

    latency_data = {}
    for model, latencies in per_model_latency.items():
        latency_data[f"llm_latency_mean_{model}"] = float(np.mean(latencies))
        latency_data[f"llm_latency_median_{model}"] = float(np.median(latencies))
        latency_data[f"llm_latency_total_{model}"] = sum(latencies)
    return latency_data


def extract_cost_data(
    all_results: T.List[SyftrEvaluationResult],
) -> T.Dict[str, float]:
    call_data = [result.llm_call_data for result in all_results]
    per_model_costs = defaultdict(list)
    for calls in call_data:
        for call in calls:
            per_model_costs[call.llm_name].append(call.cost)
    total_costs_per_model = {
        f"llm_cost_total_{model}": sum(costs)
        for model, costs in per_model_costs.items()
    }
    run_costs = [sum(call.cost for call in calls) for calls in call_data]
    return {
        "llm_cost_total": sum(run_costs),
        "llm_cost_min": float(np.min(run_costs)),
        "llm_cost_max": float(np.max(run_costs)),
        "llm_cost_mean": float(np.mean(run_costs)),
        "llm_cost_median": float(np.median(run_costs)),
        **total_costs_per_model,
    }


def extract_token_data(
    all_results: T.List[SyftrEvaluationResult],
) -> T.Dict[str, float]:
    call_data = [result.llm_call_data for result in all_results]
    per_model_input_tokens = defaultdict(list)
    for calls in call_data:
        for call in calls:
            per_model_input_tokens[call.llm_name].append(call.input_tokens)
    total_input_tokens_per_model = {
        f"llm_input_tokens_total_{model}": sum(input_tokens)
        for model, input_tokens in per_model_input_tokens.items()
    }
    run_input_tokens = [sum(call.input_tokens for call in calls) for calls in call_data]

    per_model_output_tokens = defaultdict(list)
    for calls in call_data:
        for call in calls:
            per_model_output_tokens[call.llm_name].append(call.output_tokens)
    total_output_tokens_per_model = {
        f"llm_output_tokens_total_{model}": sum(output_tokens)
        for model, output_tokens in per_model_output_tokens.items()
    }
    run_output_tokens = [
        sum(call.output_tokens for call in calls) for calls in call_data
    ]
    return {
        "llm_input_tokens_total": sum(run_input_tokens),
        "llm_input_tokens_min": float(np.min(run_input_tokens)),
        "llm_input_tokens_max": float(np.max(run_input_tokens)),
        "llm_input_tokens_mean": float(np.mean(run_input_tokens)),
        "llm_input_tokens_median": float(np.median(run_input_tokens)),
        "llm_output_tokens_total": sum(run_output_tokens),
        "llm_output_tokens_min": float(np.min(run_output_tokens)),
        "llm_output_tokens_max": float(np.max(run_output_tokens)),
        "llm_output_tokens_mean": float(np.mean(run_output_tokens)),
        "llm_output_tokens_median": float(np.median(run_output_tokens)),
        **total_input_tokens_per_model,
        **total_output_tokens_per_model,
    }


def extract_judge_results(
    all_results: T.List[SyftrEvaluationResult], include_responses: bool = False
) -> T.Dict:
    """Extract result data from judge evaluations."""
    assert all(result.qa_pair is not None for result in all_results)

    labels = np.array([result.qa_pair.label for result in all_results])  # type: ignore
    judge_responses = np.array([])
    for result, label in zip(all_results, labels):
        if result.passing is None:
            np.append(judge_responses, None)
        # Recover judge responses (as 0 or 1) from answer key + result.passing
        judge_response = label if result.passing else int(not bool(label))
        judge_responses = np.append(judge_responses, judge_response)

    # Clean out None's (eg. from failed evals)
    labels = np.array(
        [
            label
            for label, response in zip(labels, judge_responses)
            if response is not None
        ]
    )
    judge_responses = np.array(
        [response for response in judge_responses if response is not None]
    )

    if len(labels) < 5:
        return {}

    pearson_r, pearson_r_p_value = stats.pearsonr(labels, judge_responses)
    spearman_r, spearman_r_p_value = stats.spearmanr(labels, judge_responses)
    kendalltau_r, kendalltau_r_p_value = stats.kendalltau(labels, judge_responses)
    cohens_kappa = metrics.cohen_kappa_score(labels, judge_responses)
    score_judge_mean = labels.mean()
    score_human_mean = judge_responses.mean()

    return {
        "pearson_r": pearson_r,
        "pearson_r_p_value": pearson_r_p_value,
        "spearman_r": spearman_r,
        "spearman_r_p_value": spearman_r_p_value,
        "kendalltau_r": kendalltau_r,
        "kendalltau_r_p_value": kendalltau_r_p_value,
        "cohens_kappa": cohens_kappa,
        "score_judge_mean": score_judge_mean,
        "score_human_mean": score_human_mean,
        "score_judge_std": judge_responses.std(),
        "score_human_std": labels.std(),
        "score_difference_mean": score_judge_mean - score_human_mean,
        "score_difference_std": (judge_responses - labels).std(),
    }



import asyncio
import typing as T

from ray.job_submission import JobStatus

from syftr.ray.submit import get_client


def derived_representer(dumper, data):
    return dumper.represent_dict({"description": data.description})


async def iter_job_logs(job_logs: T.AsyncIterable):
    async for lines in job_logs:
        print(lines, end="")


async def iter_all_job_logs(tailers: T.List[T.AsyncIterable]):
    log_iters = [iter_job_logs(tailer) for tailer in tailers]
    await asyncio.gather(*log_iters)


def attach_logs(
    prefix: str = "<doesntmatch>",
    suffix: str = "<doesntmatch>",
    substring: str = "<doesntmatch>",
):
    """Connect to all jobs that match the given prefix, suffix, or substring"""
    client = get_client()
    job_details = client.list_jobs()
    jobs_to_tail = [
        job
        for job in job_details
        if (
            job.submission_id is not None
            and (
                job.submission_id.startswith(prefix)
                or job.submission_id.endwith(suffix)
                or substring in job.submission_id
            )
            and job.status
            not in {JobStatus.STOPPED, JobStatus.SUCCEEDED, JobStatus.FAILED}
        )
    ]
    log_tailers = [client.tail_job_logs(job.job_id) for job in jobs_to_tail]
    asyncio.run(iter_all_job_logs(log_tailers))



import getpass
import os

import ray

from syftr.configuration import cfg
from syftr.logger import logger


def ray_init(force_remote: bool = False):
    if ray.is_initialized():
        logger.warning(
            "Using existing ray client with address '%s'", ray.client().address
        )
    else:
        address = cfg.ray.remote_endpoint if force_remote else None

        if address is None:
            username = getpass.getuser()
            ray_tmpdir = f"/tmp/ray_{username}"
            logger.info(
                "Using local ray client with temporary directory '%s'", ray_tmpdir
            )
            os.environ["RAY_TMPDIR"] = ray_tmpdir

        ray.init(
            address=address,
            logging_level=cfg.logging.level,
        )



import csv
import json
import logging
import numbers
import textwrap
import traceback
import typing as T
from pathlib import Path

import numpy as np
import pandas as pd
from filelock import FileLock
from tabulate import tabulate

from syftr.configuration import cfg
from syftr.logger import logger


def text_to_bool(text):
    true_values = {"true", "1", "yes", "on", "t"}
    false_values = {"false", "0", "no", "off", "f"}

    normalized_text = text.strip().lower()

    if normalized_text in true_values:
        return True
    elif normalized_text in false_values:
        return False
    raise ValueError(f"Invalid text input for boolean conversion: '{text}'")


def format_value(value, max_width=60):
    if isinstance(value, str):
        return "\n".join(textwrap.wrap(value, max_width))
    return value


def append_dict_to_csv(file_path: Path, data: dict):
    file_path = Path(file_path).resolve()
    lock = FileLock(file_path.with_suffix(".lock"))
    with lock:
        logger.info("Appending data to file: %s", file_path)
        file_exists = file_path.exists()
        with file_path.open(mode="a", newline="") as file:
            writer = csv.DictWriter(file, fieldnames=data.keys())
            if not file_exists:
                writer.writeheader()
            writer.writerow(data)


def debug(
    key_value_pairs: dict, max_width=60, headers=None, tablefmt="grid", file_path=None
):
    headers = headers or ["Parameter", "Value"]
    if cfg.logging.level > logging.DEBUG:
        return
    table = [
        [str(k), format_value(str(v), max_width=max_width)]
        for k, v in sorted(key_value_pairs.items())
    ]
    print(tabulate(table, headers=headers, tablefmt=tablefmt))
    if file_path:
        append_dict_to_csv(file_path, key_value_pairs)


def get_baselines_from_trials(df_trials: pd.DataFrame) -> T.List[T.Dict[str, T.Any]]:
    return list(df_trials["user_attrs_flow"].apply(json.loads).values)


def get_flows_from_trials(df_trials: pd.DataFrame) -> T.List[T.Dict[str, T.Any]]:
    return list(df_trials["user_attrs_flow"].apply(json.loads).values)


def is_within_range(value, min_val, max_val, step, tolerance=1e-9):
    if not (min_val <= value <= max_val):
        return False
    diff = value - min_val
    remainder = diff % step
    return remainder < tolerance or abs(remainder - step) < tolerance


def get_exception_report(
    exception: T.Union[
        BaseException,
        BaseExceptionGroup,
    ],
) -> str:
    e = exception
    ex_type = type(e).__name__
    ex_message = str(e)
    ex_traceback = "".join(traceback.format_exception(type(e), e, e.__traceback__))
    return f"Type: {ex_type}, Message: {ex_message}, Traceback: {ex_traceback}"


def is_numeric(x):
    return isinstance(x, numbers.Number) or (
        isinstance(x, np.generic) and np.issubdtype(x, np.number)
    )


def get_unique_bools(df: pd.DataFrame, col: str) -> T.List[bool]:
    param = "params_" + col
    if param not in df.columns:
        return []
    return [value for value in df[param].unique() if isinstance(value, bool)]


def get_unique_strings(df: pd.DataFrame, col: str) -> T.List[str]:
    param = "params_" + col
    if param not in df.columns:
        return []
    return [value for value in df[param].unique() if isinstance(value, str)]


def get_unique_ints(df: pd.DataFrame, col: str) -> T.List[int]:
    param = "params_" + col
    if param not in df.columns:
        return []
    values = [
        int(round(value, 0))
        for value in df[param].unique()
        if is_numeric(value) and not pd.isna(value)
    ]
    return list(set(values))


def get_unique_floats(df: pd.DataFrame, col: str, ndigits: int) -> T.List[float]:
    param = "params_" + col
    if param not in df.columns:
        return []
    values = [
        round(float(value), ndigits)
        for value in df[param].unique()
        if is_numeric(value) and not np.isnan(value)
    ]
    return list(set(values))


def get_min_int(df: pd.DataFrame, col: str, default: int) -> int:
    ints = get_unique_ints(df, col)
    if not ints:
        return default
    return min(ints)


def get_max_int(df: pd.DataFrame, col: str, default: int) -> int:
    ints = get_unique_ints(df, col)
    if not ints:
        return default
    return max(ints)


def get_min_float(df: pd.DataFrame, col: str, default: float, ndigits: int) -> float:
    floats = get_unique_floats(df, col, ndigits)
    if not floats:
        return default
    return min(floats)


def get_max_float(df: pd.DataFrame, col: str, default: float, ndigits: int) -> float:
    floats = get_unique_floats(df, col, ndigits)
    if not floats:
        return default
    return max(floats)



import json
import time
import typing as T

from llama_index.core.bridge.pydantic import PrivateAttr

from syftr.logger import logger
from syftr.studies import TimeoutConfig


class EmbeddingPreemptiveTimeoutError(Exception):
    """Embeddings expected to go over time."""

    def __init__(self, metadata):
        message = f"Embeddings expected to exceed max time limit: {json.dumps(metadata, sort_keys=True)}"
        super().__init__(message)


class EmbeddingTimeoutMixin:
    _chunks_processed: T.Any = PrivateAttr()
    _chunks_remaining: T.Any = PrivateAttr()
    _done_building_index: T.Any = PrivateAttr()
    _seconds_per_chunk: T.Any = PrivateAttr()
    _time_remaining: T.Any = PrivateAttr()
    _time_start: T.Any = PrivateAttr()
    _timeout_config: T.Any = PrivateAttr()
    _total_chunks: T.Any = PrivateAttr()
    _total_time: T.Any = PrivateAttr()
    _warning_logged: T.Any = PrivateAttr()

    def __init__(
        self,
        *args,
        timeout_config: TimeoutConfig = TimeoutConfig(),
        total_chunks: int = 0,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self._timeout_config = timeout_config
        self._total_chunks = total_chunks

        self._chunks_processed = 0
        self._time_start = None
        self._warning_logged = False

    def reset_timeouts(
        self,
        disable_timeouts: bool = False,
        total_chunks: int | None = None,
        timeout_config: TimeoutConfig | None = None,
    ):
        """Enable timeouts to be reset or disabled.

        Useful when index building has finished and we are moving to evaluation,
        or when reusing for a separate index.
        """
        self._chunks_processed = 0
        self._time_start = None
        self._done_building_index = bool(disable_timeouts)
        self._warning_logged = False

        if total_chunks is not None:
            self._total_chunks = total_chunks

        if timeout_config is not None:
            self._timeout_config = timeout_config

    @property
    def metadata(self) -> T.Dict[str, T.Any]:
        return {
            "_chunks_processed": self._chunks_processed,
            "_chunks_remaining": self._chunks_remaining,
            "_done_building_index": self._done_building_index,
            "_seconds_per_chunk": self._seconds_per_chunk,
            "_time_remaining": self._time_remaining,
            "_time_start": self._time_start,
            "_timeout_config": self._timeout_config.dict(),
            "_total_chunks": self._total_chunks,
            "_total_time": self._total_time,
            "_warning_logged": self._warning_logged,
        }

    def _check_remaining_time(self):
        if self._total_chunks == 0:
            return

        if self._done_building_index:
            return

        min_chunks_met = (
            self._chunks_processed
            > self._timeout_config.embedding_min_chunks_to_process
        )
        min_time_spent = (
            self._total_time > self._timeout_config.embedding_min_time_to_process
        )

        if min_chunks_met or min_time_spent:
            self._seconds_per_chunk = self._total_time / self._chunks_processed
            self._chunks_remaining = self._total_chunks - self._chunks_processed
            self._time_remaining = self._seconds_per_chunk * self._chunks_remaining
            if self._chunks_processed % 3200 == 0:
                logger.info(
                    "%s/%s chunks processed at %s it/s after %s s. Estimated time remaining is %s s",
                    self._chunks_processed,
                    self._total_chunks,
                    self._total_time,
                    self._chunks_processed / self._total_time,
                    self._time_remaining,
                )

            if (
                self._total_time + self._time_remaining
                > self._timeout_config.embedding_max_time
            ):
                if self._timeout_config.embedding_timeout_active:
                    raise EmbeddingPreemptiveTimeoutError(self.metadata)
                if not self._warning_logged:
                    logger.warning(
                        "Not raising embedding timeout because it is deactivated"
                    )
                    self._warning_logged = True

    def _embed(self, sentences: T.List[str], *args, **kwargs) -> T.List[T.List[float]]:
        if self._time_start is None:
            self._time_start = time.time()

        result = super()._embed(sentences, *args, **kwargs)  # type: ignore

        self._total_time = time.time() - self._time_start
        self._chunks_processed += len(sentences)
        self._check_remaining_time()
        return result

    def _compute_embeddings(
        self, texts: T.List[str], prefix: str | None = None
    ) -> T.List[T.List[float]]:
        if self._time_start is None:
            self._time_start = time.time()

        result = super()._compute_embeddings(texts, prefix=prefix)  # type: ignore

        self._total_time = time.time() - self._time_start
        self._chunks_processed += len(texts)
        self._check_remaining_time()
        return result



import argparse
import asyncio
import getpass
from datetime import datetime
from pathlib import Path
from typing import Any, Dict

import git
import pytz
from ray.job_submission import JobSubmissionClient

from syftr.configuration import cfg
from syftr.logger import logger
from syftr.optimization import user_confirm_delete
from syftr.ray.runtime_env import get_runtime_env
from syftr.ray.utils import ray_init
from syftr.studies import AgentStudyConfig, StudyConfig, get_default_study_name


def get_client() -> JobSubmissionClient:
    address = cfg.ray.local_endpoint if cfg.ray.local else cfg.ray.remote_endpoint
    logger.info("Connecting to Ray job submission client at `%s`", address)
    if cfg.ray.local:
        ray_init()
    return JobSubmissionClient(address)


def _get_metadata(study_config: StudyConfig) -> Dict[str, Any]:
    try:
        repo = git.Repo(cfg.paths.root_dir)
        sha = repo.head.commit.hexsha
        short_sha = repo.git.rev_parse(sha, short=True)
    except git.exc.InvalidGitRepositoryError:
        # We are not in a git repo, syftr is used as a library,
        short_sha = None
    metadata = {
        "study_name": study_config.name,
        "submitted_by": getpass.getuser(),
        "study_config": study_config.model_dump_json(),
        "config": cfg.model_dump_json(),
    }
    if short_sha:
        metadata["git_short_sha"] = short_sha
    return metadata


def _get_submission_id(metadata: Dict[str, Any]):
    utc = pytz.utc
    current_time_utc = datetime.now(utc)
    t = current_time_utc.strftime("%Y-%m-%d-%H:%M:%S")
    s = metadata["study_name"]
    u = metadata["submitted_by"]
    submission_id = f"{t}--{u}--{s}"
    short_sha = metadata.get("git_short_sha")
    if short_sha:
        submission_id += f"--{short_sha}"
    return submission_id


def start_study(
    client: JobSubmissionClient,
    study_config_file: Path,
    study_config: StudyConfig,
    delete_confirmed: bool = False,
    agentic: bool = False,
) -> str:
    metadata = _get_metadata(study_config)
    submission_id = _get_submission_id(metadata)
    runtime_env = get_runtime_env(study_config_file, delete_confirmed)

    if not agentic:
        entrypoint = (
            f"python -m syftr.tuner.qa_tuner --study-config {study_config_file.name}"
        )
    else:
        entrypoint = (
            f"python -m syftr.tuner.agent_tuner --study-config {study_config_file.name}"
        )

    if not cfg.ray.local:
        logger.info(
            "Submitting job to remote Ray cluster at `%s`", cfg.ray.remote_endpoint
        )
    else:
        logger.info("Submitting job to local Ray instance")

    job_id = client.submit_job(
        submission_id=submission_id,
        entrypoint=entrypoint,
        runtime_env=runtime_env,
        metadata=metadata,
    )
    return job_id


def tail(client: JobSubmissionClient, job_id: str):
    async def _tail():
        async for line in client.tail_job_logs(job_id):
            print(line)

    try:
        asyncio.run(_tail())
    except Exception:
        logger.exception("Exception raised during execution of job `%s`", job_id)
    except KeyboardInterrupt:
        logger.warning("KeyboardInterrupt received for job %s", job_id)
    finally:
        logger.info("Done tailing logs from job `%s`", job_id)
        info = client.get_job_info(job_id)
        logger.info(info)
        if not cfg.ray.local:
            stop_command = f"ray job stop --address {cfg.ray.remote_endpoint} {job_id}"
            logger.info("Client disconnected but job may still be running.")
            logger.info("Job status: %s - %s", job_id, info.status)
            logger.info("To stop the job, run `%s`", stop_command)


def main():
    parser = argparse.ArgumentParser()
    default = get_default_study_name()
    parser.add_argument(
        "--study-config",
        help=f"Path to study config yaml (default: {default})",
        default=default,
    )
    parser.add_argument(
        "--remote",
        help="Use remote Ray cluster",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--agent",
        help="Whether to run agent optimization",
        default=False,
        action="store_true",
    )
    args = parser.parse_args()

    study_config_file = Path(args.study_config)
    if not args.agent:
        study_config = StudyConfig.from_file(study_config_file)
    else:
        study_config = AgentStudyConfig.from_file(study_config_file)
    delete_confirmed = user_confirm_delete(study_config)

    cfg.ray.local = False if args.remote else cfg.ray.local
    client = get_client()
    logger.info(
        "Starting study `%s` from file `%s`",
        study_config.name,
        study_config_file.as_posix(),
    )
    job_id = start_study(
        client,
        study_config_file,
        study_config,
        delete_confirmed=delete_confirmed,
        agentic=args.agent,
    )
    logger.info("Started study `%s` with job_id `%s`", study_config.name, job_id)
    tail(client, job_id)


if __name__ == "__main__":
    main()



from syftr.studies import SearchSpace
from syftr.validation import (
    has_valid_additional_context_params,
    has_valid_base_rag_params,
    has_valid_few_shot_params,
    has_valid_hyde_params,
    has_valid_minimum_params,
    has_valid_rag_embedding_model_params,
    has_valid_rag_fusion_params,
    has_valid_rag_reranker_params,
    has_valid_splitter_params,
    has_valid_unique_params,
)


def test_params():
    """Simple unit test that must be updated whenever params update.

    Serves as tested documentation
    """
    ss = SearchSpace()
    params = ss.param_names()
    for param in sorted(params):
        print(f'"{param}",')
    assert sorted(params) == sorted(
        [
            "additional_context_enabled",
            "additional_context_num_nodes",
            "coa_enable_calculator",
            "critique_agent_llm_name",
            "critique_agent_llm_temperature",
            "critique_agent_llm_top_p",
            "critique_agent_llm_use_reasoning",
            "few_shot_embedding_model",
            "few_shot_enabled",
            "few_shot_top_k",
            "hyde_enabled",
            "hyde_llm_name",
            "hyde_llm_temperature",
            "hyde_llm_top_p",
            "hyde_llm_use_reasoning",
            "lats_max_rollouts",
            "lats_num_expansions",
            "rag_embedding_model",
            "rag_fusion_mode",
            "rag_hybrid_bm25_weight",
            "rag_method",
            "rag_mode",
            "rag_query_decomposition_enabled",
            "rag_query_decomposition_llm_name",
            "rag_query_decomposition_llm_temperature",
            "rag_query_decomposition_llm_top_p",
            "rag_query_decomposition_llm_use_reasoning",
            "rag_query_decomposition_num_queries",
            "rag_top_k",
            "reflection_agent_llm_name",
            "reflection_agent_llm_temperature",
            "reflection_agent_llm_top_p",
            "reflection_agent_llm_use_reasoning",
            "reranker_enabled",
            "reranker_llm_name",
            "reranker_llm_temperature",
            "reranker_llm_top_p",
            "reranker_llm_use_reasoning",
            "reranker_top_k",
            "response_synthesizer_llm_name",
            "response_synthesizer_llm_temperature",
            "response_synthesizer_llm_top_p",
            "response_synthesizer_llm_use_reasoning",
            "splitter_chunk_exp",
            "splitter_chunk_overlap_frac",
            "splitter_method",
            "subquestion_engine_llm_name",
            "subquestion_engine_llm_temperature",
            "subquestion_engine_llm_top_p",
            "subquestion_engine_llm_use_reasoning",
            "subquestion_response_synthesizer_llm_name",
            "subquestion_response_synthesizer_llm_temperature",
            "subquestion_response_synthesizer_llm_top_p",
            "subquestion_response_synthesizer_llm_use_reasoning",
            "template_name",
        ]
    )


def test_minimum_params(search_space_and_trial_params):
    assert has_valid_minimum_params(*search_space_and_trial_params), (
        "Minimum parameters required are invalid"
    )


def test_few_shot_params(search_space_and_trial_params):
    assert has_valid_few_shot_params(*search_space_and_trial_params), (
        "Few shot parameters are invalid"
    )


def test_base_rag_params(search_space_and_trial_params):
    assert has_valid_base_rag_params(*search_space_and_trial_params), (
        "Base RAG parameters are invalid"
    )


def test_rag_embedding_model_params(search_space_and_trial_params):
    assert has_valid_rag_embedding_model_params(*search_space_and_trial_params), (
        "RAG embedding model parameters are invalid"
    )


def test_fusion_params(search_space_and_trial_params):
    assert has_valid_rag_fusion_params(*search_space_and_trial_params), (
        "Fusion parameters are invalid"
    )


def test_reranker_params(search_space_and_trial_params):
    assert has_valid_rag_reranker_params(*search_space_and_trial_params), (
        "Reranker parameters are invalid"
    )


def test_hyde_params(search_space_and_trial_params):
    assert has_valid_hyde_params(*search_space_and_trial_params), (
        "Hyde params are invalid"
    )


def test_splitter_params(search_space_and_trial_params):
    assert has_valid_splitter_params(*search_space_and_trial_params), (
        "Splitter params are invalid"
    )


def test_out_of_distribution_rag_top_k(search_space_and_trial_params):
    search_space_and_trial_params[1]["rag_top_k"] = int(1e6)
    is_valid = has_valid_rag_embedding_model_params(*search_space_and_trial_params)
    assert (
        is_valid
        if search_space_and_trial_params[1]["rag_mode"] == "no_rag"
        else not is_valid
    ), "Out-of-distribution check for 'rag_top_k' failed"


def test_additional_context_params(search_space_and_trial_params):
    assert has_valid_additional_context_params(*search_space_and_trial_params), (
        "Additional context params are invalid"
    )


def test_has_valid_unique_params(
    search_space_and_trial_params,
):
    assert has_valid_unique_params(
        *search_space_and_trial_params,
        "critique_rag_agent",
        [
            "critique_agent_llm_name",
            "reflection_agent_llm_name",
        ],
    ), (
        f"Parameters for RAG mode {search_space_and_trial_params[1]['rag_mode']} are in conflict with CritiqueRAG"
    )

    assert has_valid_unique_params(
        *search_space_and_trial_params,
        "lats_rag_agent",
        [
            "lats_max_rollouts",
            "lats_num_expansions",
        ],
    ), (
        f"Parameters for RAG mode {search_space_and_trial_params[1]['rag_mode']} are in conflict with LATS"
    )



import pathlib
from pathlib import Path
from unittest.mock import MagicMock, patch

import optuna
import pandas as pd
import pytest

from syftr import api
from syftr.configuration import cfg
from syftr.studies import StudyConfig

# Create a mock study object that contains 10 trials of actual data
MOCK_STUDY = MagicMock(spec=optuna.Study)
MOCK_STUDY.study_name = "example-dr-docs"
TRIALS_DF_PATH = pathlib.Path(__file__).parent / "mock_trials_df.pkl"
MOCK_TRIALS_DF = pd.read_pickle(TRIALS_DF_PATH)
MOCK_STUDY.trials_dataframe.return_value = MOCK_TRIALS_DF

EXAMPLE_STUDY_CONFIG_PATH = Path(cfg.paths.studies_dir / "example-dr-docs.yaml")
NON_EXISTANT_STUDY_CONFIG_PATH = Path(
    cfg.paths.test_studies_dir / "hotpot-toy-non-existent.yaml"
)


# This fixture mocks the optuna storage so that no actual database connection is needed
@pytest.fixture(autouse=True)
def mock_optuna_storage():
    with patch("syftr.configuration.Database.get_optuna_storage") as mock_storage:
        mock_storage.return_value = None
        yield


def test_init_no_path():
    study_config_path = Path(EXAMPLE_STUDY_CONFIG_PATH)
    study = api.Study(StudyConfig.from_file(study_config_path))
    assert study.study_path == study_config_path


def test_pareto_df():
    with patch("syftr.api.optuna.load_study", return_value=MOCK_STUDY):
        study = api.Study.from_file(EXAMPLE_STUDY_CONFIG_PATH)
        pareto_df = study.pareto_df
    assert pareto_df is not None
    assert isinstance(pareto_df, pd.DataFrame)
    assert not pareto_df.empty
    assert "user_attrs_flow" in pareto_df.columns
    assert "values_0" in pareto_df.columns
    assert "values_1" in pareto_df.columns


def test_flows_df_returns_dataframe():
    with patch("syftr.api.optuna.load_study", return_value=MOCK_STUDY):
        study = api.Study.from_file(EXAMPLE_STUDY_CONFIG_PATH)
        df = study.flows_df
    assert isinstance(df, pd.DataFrame)
    assert not df.empty
    assert "user_attrs_flow" in df.columns
    assert "values_0" in df.columns
    assert "values_1" in df.columns


def test_pareto_df_study_does_not_exist():
    study = api.Study.from_file(NON_EXISTANT_STUDY_CONFIG_PATH)
    with pytest.raises(api.SyftrUserAPIError) as exc:
        study.pareto_df
    assert exc.value.args[0].startswith("Cannot find this study in the database:")


def test_pareto_flows():
    with patch("syftr.api.optuna.load_study", return_value=MOCK_STUDY):
        study = api.Study.from_file(EXAMPLE_STUDY_CONFIG_PATH)
        pareto_flows = study.pareto_flows
    assert pareto_flows
    assert all("llm_cost_mean" in flow["metrics"] for flow in pareto_flows)
    assert all("accuracy" in flow["metrics"] for flow in pareto_flows)


def test_pareto_flows_study_does_not_exist():
    study = api.Study.from_file(NON_EXISTANT_STUDY_CONFIG_PATH)
    with pytest.raises(api.SyftrUserAPIError) as exc:
        study.pareto_flows
    assert exc.value.args[0].startswith("Cannot find this study in the database")


def test_knee_point():
    with patch("syftr.api.optuna.load_study", return_value=MOCK_STUDY):
        study = api.Study.from_file(EXAMPLE_STUDY_CONFIG_PATH)
        with pytest.raises(api.SyftrUserAPIError) as exc:
            study.knee_point
    assert (
        exc.value.args[0]
        == "Not enough points in the Pareto front to find a knee point."
    )


def test_knee_point_study_does_not_exist():
    study = api.Study.from_file(NON_EXISTANT_STUDY_CONFIG_PATH)
    with pytest.raises(api.SyftrUserAPIError) as exc:
        study.knee_point
    assert exc.value.args[0].startswith("Cannot find this study in the database")


def test_status_completed():
    with patch("syftr.api.optuna.load_study", return_value=MOCK_STUDY):
        study = api.Study.from_file(EXAMPLE_STUDY_CONFIG_PATH)
        status = study.status["job_status"]
    assert status == api.SyftrStudyStatus.COMPLETED


def test_status_non_existent():
    study = api.Study.from_file(NON_EXISTANT_STUDY_CONFIG_PATH)
    assert study.status["job_status"] == api.SyftrStudyStatus.INITIALIZED


def test_get_study_non_existent_in_db():
    with pytest.raises(api.SyftrUserAPIError) as exc:
        api.Study.from_db("non_existent_study")
    assert exc.value.args[0] == "Cannot find study non_existent_study in the database."


def test_from_file_study():
    study = api.Study.from_file(EXAMPLE_STUDY_CONFIG_PATH)
    assert study
    assert isinstance(study, api.Study)



import os
from typing import Optional

import boto3
from botocore.exceptions import ClientError
from dotenv import load_dotenv, set_key

from syftr.configuration import REPO_ROOT, S3_TIMEOUT, cfg
from syftr.logger import logger
from syftr.utils.locks import distributed_lock


def get_parameter_value(client, name, decrypt) -> str:
    logger.info(f"Loading parameter value '{name}' with client")
    result = client.get_parameter(Name=name, WithDecryption=decrypt)
    return result["Parameter"]["Value"]


def get_aws_keypair():
    """Retrieve AWS keypair for octo-syftr user."""
    load_dotenv(REPO_ROOT / ".env")
    access_key = os.getenv("AWS_ACCESS_KEY_ID")
    secret_key = os.getenv("AWS_SECRET_ACCESS_KEY")
    if access_key and secret_key:
        return {
            "AWS_ACCESS_KEY_ID": access_key,
            "AWS_SECRET_ACCESS_KEY": secret_key,
            "AWS_DEFAULT_REGION": cfg.aws.region,
        }
    client = boto3.client("ssm")
    access_key = access_key or get_parameter_value(
        client, cfg.aws.access_key_ssm_path, True
    )
    secret_key = secret_key or get_parameter_value(
        client, cfg.aws.secret_key_ssm_path, True
    )
    return {
        "AWS_ACCESS_KEY_ID": access_key,
        "AWS_SECRET_ACCESS_KEY": secret_key,
        "AWS_DEFAULT_REGION": cfg.aws.region,
    }


def set_aws_key_pair(key_pair: dict):
    for k, v in key_pair.items():
        assert k in ["AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY", "AWS_DEFAULT_REGION"]
        set_key(REPO_ROOT / ".env", k, v)


def check_file_exists_s3(object_key):
    s3 = boto3.client("s3")
    try:
        s3.head_object(Bucket=cfg.storage.cache_bucket, Key=object_key)
        return True
    except ClientError as e:
        if e.response["Error"]["Code"] == "404":
            return False
        raise


def get_file_from_s3(object_key) -> Optional[str]:
    s3 = boto3.client("s3")

    with distributed_lock(
        f"{cfg.storage.cache_bucket}/{object_key}", timeout_s=S3_TIMEOUT
    ):
        if check_file_exists_s3(object_key):
            response = s3.get_object(Bucket=cfg.storage.cache_bucket, Key=object_key)
            # Read the file content into memory as a byte string
            file_content = response["Body"].read()
            return file_content

    logger.debug(
        f"File does not exist in bucket '{cfg.storage.cache_bucket}' with key '{object_key}'."
    )
    return None


def delete_file_from_s3(object_key) -> None:
    s3 = boto3.client("s3")

    with distributed_lock(
        f"{cfg.storage.cache_bucket}/{object_key}", timeout_s=S3_TIMEOUT
    ):
        if check_file_exists_s3(object_key):
            s3.delete_object(Bucket=cfg.storage.cache_bucket, Key=object_key)


if __name__ == "__main__":
    key_pair = get_aws_keypair()
    set_aws_key_pair(key_pair)



import re
import typing as T

import pandas as pd
from openinference.instrumentation.llama_index import LlamaIndexInstrumentor
from opentelemetry import trace as trace_api
from opentelemetry.sdk import trace as trace_sdk
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import (
    SimpleSpanProcessor,
    SpanExporter,
    SpanExportResult,
)


class InMemorySpanExporter(SpanExporter):
    def __init__(self):
        self.finished_spans = []

    def export(self, spans: T.Sequence[ReadableSpan]) -> SpanExportResult:
        self.finished_spans.extend(spans)
        return SpanExportResult.SUCCESS

    def shutdown(self):
        pass


def get_span_exporter() -> InMemorySpanExporter:
    resource = Resource(attributes={})
    tracer_provider = trace_sdk.TracerProvider(resource=resource)
    span_exporter = InMemorySpanExporter()
    span_processor = SimpleSpanProcessor(span_exporter)
    tracer_provider.add_span_processor(span_processor)
    trace_api.set_tracer_provider(tracer_provider=tracer_provider)
    LlamaIndexInstrumentor().instrument(tracer_provider=tracer_provider)
    return span_exporter


def span_to_dataframe(span: ReadableSpan):
    data = {
        "span_name": [span.name],
        "span_start_time": [span.start_time],
        "span_end_time": [span.end_time],
        "span_trace_id": [span.context.trace_id],
    }

    if span.attributes:
        for key, value in span.attributes.items():
            column = f"attribute_{key}"
            if re.search(r".*\.\d+\..*", key):
                tmp = re.sub(r"\.\d+\.", ".", key)
                column = f"attribute_{tmp}"
            data[column] = [value]

    return pd.DataFrame(data)


def spans_to_dataframe(spans: T.Sequence[ReadableSpan]):
    dataframes = [span_to_dataframe(span) for span in spans]
    return pd.concat(dataframes, ignore_index=True)


def _set_tracing_data(df: pd.DataFrame, results: T.Dict[str, T.Any]):
    prefix = "trace__"
    num_decimals = 2

    results[prefix + "num_trace_ids"] = len(df["span_trace_id"].unique())

    df["span_duration"] = df["span_end_time"] - df["span_start_time"]
    trace_avg_durations = df.groupby("span_name")["span_duration"].mean().to_dict()
    trace_avg_durations = {
        f"{prefix}{key.replace('__', '')}__duration_mean": round(value, num_decimals)
        for key, value in trace_avg_durations.items()
    }
    results.update(trace_avg_durations)

    trace_total_durations = df.groupby("span_name")["span_duration"].sum().to_dict()
    trace_total_durations = {
        f"{prefix}{key.replace('__', '')}__duration_total": value
        for key, value in trace_total_durations.items()
    }
    results.update(trace_total_durations)

    df["input_length"] = df["attribute_input.value"].apply(
        lambda x: len(x) if isinstance(x, str) else 0
    )
    df["output_lenght"] = df["attribute_output.value"].apply(
        lambda x: len(x) if isinstance(x, str) else 0
    )

    trace_avg_input_lengths = df.groupby("span_name")["input_length"].mean().to_dict()
    trace_avg_input_lengths = {
        f"{prefix}{key.replace('__', '')}__input_length_mean": round(
            value, num_decimals
        )
        for key, value in trace_avg_input_lengths.items()
    }
    results.update(trace_avg_input_lengths)

    trace_avg_ouptput_lengths = (
        df.groupby("span_name")["output_lenght"].mean().to_dict()
    )
    trace_avg_ouptput_lengths = {
        f"{prefix}{key.replace('__', '')}__output_length_mean": round(
            value, num_decimals
        )
        for key, value in trace_avg_ouptput_lengths.items()
    }
    results.update(trace_avg_ouptput_lengths)


def set_tracing_metrics(
    span_exporter: InMemorySpanExporter, results: T.Dict[str, T.Any]
):
    spans = span_exporter.finished_spans
    if not spans:
        return

    df = spans_to_dataframe(spans)
    _set_tracing_data(df, results)



import json
import typing as T
from dataclasses import dataclass
from pathlib import Path

from syftr.configuration import cfg


@dataclass
class SOTA:
    name: str
    accuracy: float
    link: str
    dataset: str
    subset: str | None = None
    test_setting: str | None = None
    eval_method: str | None = None
    method_name: str | None = None


def load_sota(file_path: Path) -> list[SOTA]:
    with open(file_path) as f:
        return [SOTA(**sota) for sota in json.load(f)]


_SOTAS: T.List[SOTA] = load_sota(cfg.paths.sota_dir / "sota_data.json")


def get_sota(study_name) -> SOTA | None:
    match study_name.lower():
        case name if "hotpot" in name:
            return next((s for s in _SOTAS if s.dataset == "HotpotQA"), None)
        case name if "financebench" in name:
            return next((s for s in _SOTAS if s.dataset == "FinanceBench"), None)
        case name if "crag" in name and "finance" in name:
            return next(
                (
                    s
                    for s in _SOTAS
                    if s.dataset == "CRAG Task3" and s.subset == "Finance"
                ),
                None,
            )
        case name if "crag" in name and "sports" in name:
            return next(
                (
                    s
                    for s in _SOTAS
                    if s.dataset == "CRAG Task3" and s.subset == "Sports"
                ),
                None,
            )
        case name if "crag" in name and "music" in name:
            return next(
                (
                    s
                    for s in _SOTAS
                    if s.dataset == "CRAG Task3" and s.subset == "Music"
                ),
                None,
            )
        case name if "crag" in name and "movie" in name:
            return next(
                (
                    s
                    for s in _SOTAS
                    if s.dataset == "CRAG Task3" and s.subset == "Movie"
                ),
                None,
            )
        case name if "crag" in name and "open" in name:
            return next(
                (s for s in _SOTAS if s.dataset == "CRAG Task3" and s.subset == "Open"),
                None,
            )
        case _:
            return None
